/*
** crampass.c - /etc/cram-md5.pwd update tool
**
** Copyright (c) 2002 Peter Eriksson <pen@lysator.liu.se>
**
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**
** This tool can be used by normal users to update the /etc/cram-md5.pwd
** CRAM MD5 password database used by the WU IMAP and POP daemons.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <crypt.h>
#include <pwd.h>
#include <signal.h>
#include <setjmp.h>

#ifndef PATH_CRAM_MD5_PWD
#define PATH_CRAM_MD5_PWD "/etc/cram-md5.pwd"
#endif

#ifndef PATH_CRAM_MD5_TMP
#define PATH_CRAM_MD5_TMP "/etc/cram-md5.tmp"
#endif

char *argv0 = "crampass.c";
int verbose = 1;
jmp_buf jenv;


void
sigint_handler(int sig)
{
    longjmp(jenv, 1);
}


void
update(char *user,
       char *pass)
{
    FILE *fp;
    int fd = -1, i = 0;
    char buf[1024], *up, *pp;


    if (pass)
    {
	if (strlen(user) + strlen(pass) + 2 > sizeof(buf))
	{
	    fprintf(stderr, "%s: Length of username and password too big.\n",
		    argv0);
	    exit(1);
	}
    }

    
    if (setjmp(jenv))
    {
	fprintf(stderr, "%s: Interrupted.\n", argv0);
	
	if (fd != -1)
	    unlink(PATH_CRAM_MD5_TMP);
	
	exit(1);
    }

    signal(SIGINT, sigint_handler);
    
    
    while ((fd = open(PATH_CRAM_MD5_TMP, O_WRONLY+O_CREAT+O_EXCL, 0600)) < 0)
    {
	if ((errno != EINTR && errno != EEXIST) || ++i > 5)
	{
	    fprintf(stderr, "%s: Error creating new password file: %s\n",
		    argv0, strerror(errno));
	    exit(1);
	}

	if (i == 2)
	    fprintf(stderr, "%s: Temporary password file busy, retrying...\n",
		    argv0);
	
	sleep(1);
    }

    if (pass)
    {
	sprintf(buf, "%s\t%s\n", user, pass);
	
	if (write(fd, buf, strlen(buf)) != strlen(buf))
	{
	    fprintf(stderr, "%s: Error writing new password file: %s\n",
		    argv0, strerror(errno));
	    unlink(PATH_CRAM_MD5_TMP);
	    exit(1);
	}
    }


    i = 0;
    
    fp = fopen(PATH_CRAM_MD5_PWD, "r");
    if (fp != NULL)
    {
	while (fgets(buf, sizeof(buf), fp))
	{
	    ++i;
	    up = strtok(buf, " \t\n\r");
	    if (!up)
	    {
		fprintf(stderr, "%s: Invalid password file entry at line %d. Ignored.\n",
			argv0, i);
	    }
	    else
	    {
		pp = strtok(NULL, " \t\n\r");
		if (!pp)
		{
		    fprintf(stderr, "%s: Invalid password file entry at line %d. Ignored.\n",
			    argv0, i);
		}
		else
		{
		    if (strcmp(up, user) != 0)
		    {
			sprintf(buf, "%s\t%s\n", up, pp);
			if (write(fd, buf, strlen(buf)) != strlen(buf))
			{
			    fprintf(stderr, "%s: Error writing new password file: %s\n",
				    argv0, strerror(errno));
			    unlink(PATH_CRAM_MD5_TMP);
			    exit(1);
			}
		    }
		}
	    }
	}

	fclose(fp);
    }
    
    rename(PATH_CRAM_MD5_TMP, PATH_CRAM_MD5_PWD);
    close(fd);
}


int
main(int argc,
     char *argv[])
{
    int i, uid = -1;
    struct passwd *pp = NULL;
    char *user = NULL;
    char *pass = NULL;
    char *cp = NULL;
    int f_remove = 0;
    

    argv0 = argv[0];
    

    uid = getuid();
    pp = getpwuid(uid);


    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	switch (argv[i][1])
	{
	  case 'r':
	    f_remove = 1;
	    break;
	    
	  case 'u':
	    if (uid == 0)
	    {
		if (argv[i][2])
		    user = argv[i]+2;
		else
		{
		    if (++i < argc)
			user = argv[i];
		    else
		    {
			fprintf(stderr, "%s: Data missing: Username for -u switch.\n",
				argv0);
			exit(1);
		    }
		}
	    }
	    else
	    {
		fprintf(stderr, "%s: Permission denied: -u switch only allowed for root.\n",
			argv0);
		exit(1);
	    }
	    break;

	  case 's':
	    verbose = 0;
	    break;
	    
	  case 'h':
	    printf("Usage: %s [-h] [-s] [-u<user>] [-r] [<password>]\n", argv0);
	    exit(0);
	    
	  default:
	    fprintf(stderr, "%s: Invalid command line option: %s\n",
		    argv0, argv[i]);
	    exit(1);
	}
	    
    if (i < argc)
	pass = argv[i++];
    
    if (i < argc)
    {
	fprintf(stderr, "%s: Extra command line arguments (ignored)\n",
		argv0);
    }

    if (!f_remove)
    {	
	if (!user)
	{
	    if (!pp)
	    {
		fprintf(stderr, "%s: You do not exist. Go away!\n",
			argv0);
		exit(1);
	    }
	    
	    user = pp->pw_name;
	}
	else
	{
	    pp = getpwnam(user);
	    if (!pp)
	    {
		fprintf(stderr, "%s: Invalid user: Does not exist: %s\n",
			argv0, user);
		exit(1);
	    }
	}
	
	if (verbose)
	    printf("Updating CRAM MD5 password for user '%s'.\n", user);
	
	if (!pass)
	    pass = getpass("Enter password: ");
	
	cp = crypt(pass, pp->pw_passwd);
	if (strcmp(cp, pp->pw_passwd) == NULL)
	{
	    fprintf(stderr, "%s: Invalid password: Password must differ from login password.\n",
		    argv0);
	    exit(1);
	}
	
	update(user, pass);
    }
    else
    {
	if (!user)
	{
	    if (!pp)
	    {
		fprintf(stderr, "%s: You do not exist. Go away!\n",
			argv0);
		exit(1);
	    }
	    
	    user = pp->pw_name;
	}
	
	if (verbose)
	    printf("Removing CRAM MD5 password for user '%s'.\n", user);
	
	update(user, NULL);
    }
    
    exit(0);
}
